from __future__ import annotations

import json
from typing import Any, Callable, Mapping, Optional, Sequence, TypeVar

from ai.backend.client.exceptions import BackendAPIError
from ai.backend.common.types import ResultSet

from .types import BaseOutputHandler, FieldSpec, PaginatedResult

_json_opts: Mapping[str, Any] = {"indent": 2}

T = TypeVar("T")


class JsonOutputHandler(BaseOutputHandler):
    def print_item(
        self,
        item: Mapping[str, Any] | None,
        fields: Sequence[FieldSpec],
    ) -> None:
        if item is None:
            print(
                json.dumps({
                    "count": 0,
                    "total_count": 0,
                    "items": [],
                })
            )
            return
        field_map = {f.field_name: f for f in fields}
        print(
            json.dumps(
                {
                    "count": 1,
                    "total_count": 1,
                    "items": [
                        {
                            field_map[k].alt_name: field_map[k].formatter.format_json(
                                v, field_map[k]
                            )
                            for k, v in item.items()
                            if k in field_map
                        },
                    ],
                },
                **_json_opts,
            )
        )

    def print_items(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
    ) -> None:
        field_map = {f.field_name: f for f in fields}
        print(
            json.dumps(
                {
                    "count": len(items),
                    "total_count": len(items),
                    "items": [
                        {
                            field_map[k].alt_name: field_map[k].formatter.format_json(
                                v, field_map[k]
                            )
                            for k, v in item.items()
                            if k in field_map
                        }
                        for item in items
                    ],
                },
                **_json_opts,
            )
        )

    def print_result_set(
        self,
        result_set: ResultSet,
    ) -> None:
        print(json.dumps(result_set))

    def print_list(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
        *,
        is_scalar: bool = False,
    ) -> None:
        if is_scalar:
            assert len(fields) == 1
            item_list = [
                {
                    fields[0].alt_name: fields[0].formatter.format_json(item, fields[0]),
                }
                for item in items
            ]
        else:
            field_map = {f.field_name: f for f in fields}
            item_list = [
                {
                    field_map[k].alt_name: field_map[k].formatter.format_json(v, field_map[k])
                    for k, v in item.items()
                    if k in field_map
                }
                for item in items
            ]
        print(
            json.dumps(
                {
                    "count": len(items),
                    "total_count": len(items),
                    "items": item_list,
                },
                **_json_opts,
            )
        )

    def print_paginated_list(
        self,
        fetch_func: Callable[[int, int], PaginatedResult],
        initial_page_offset: int,
        page_size: int = None,
        plain=False,
    ) -> None:
        page_size = page_size or 20
        result = fetch_func(initial_page_offset, page_size)
        field_map = {f.field_name: f for f in result.fields}
        print(
            json.dumps(
                {
                    "count": len(result.items),
                    "total_count": result.total_count,
                    "items": [
                        {
                            field_map[k].alt_name: field_map[k].formatter.format_json(
                                v, field_map[k]
                            )
                            for k, v in item.items()
                            if k in field_map
                        }
                        for item in result.items
                    ],
                },
                **_json_opts,
            )
        )

    def print_mutation_result(
        self,
        item: Mapping[str, Any],
        item_name: Optional[str] = None,
        action_name: Optional[str] = None,
        extra_info: Mapping = {},
    ) -> None:
        data = {
            "ok": item.get("ok", False),
            "msg": item.get("msg", "Failed"),
            **extra_info,
        }
        if item_name is not None and item_name in item:
            data = {
                **data,
                item_name: {k: v for k, v in item[item_name].items()},
            }
        print(
            json.dumps(
                data,
                **_json_opts,
            )
        )

    def print_mutation_error(
        self,
        error: Optional[Exception] = None,
        msg: str = "Failed",
        item_name: Optional[str] = None,
        action_name: Optional[str] = None,
        extra_info: Mapping = {},
    ) -> None:
        data = {
            "ok": False,
            "msg": msg,
            "item_name": item_name,
            "action_name": action_name,
            **extra_info,
        }
        if error is not None:
            data["error"] = str(error)
        print(
            json.dumps(
                data,
                **_json_opts,
            )
        )

    def print_error(
        self,
        error: Exception,
    ) -> None:
        match error:
            case BackendAPIError():
                print(
                    json.dumps(
                        {
                            "error": error.data["title"],
                            "api": {
                                "status": error.status,
                                "reason": error.reason,
                                **error.data,
                            },
                        },
                        **_json_opts,
                    )
                )
            case _:
                print(
                    json.dumps(
                        {
                            "error": str(error),
                        },
                        **_json_opts,
                    )
                )

    def print_fail(
        self,
        message: str,
    ) -> None:
        print(
            json.dumps(
                {
                    "error": message,
                },
                **_json_opts,
            )
        )



import textwrap
from typing import Any, Iterable, Sequence

from ...cli.types import Undefined, undefined
from ..output.fields import domain_fields
from ..output.types import FieldSpec
from ..session import api_session
from ..types import set_if_set
from .base import BaseFunction, api_function, resolve_fields

__all__ = ("Domain",)

_default_list_fields = (
    domain_fields["name"],
    domain_fields["description"],
    domain_fields["is_active"],
    domain_fields["created_at"],
    domain_fields["total_resource_slots"],
    domain_fields["allowed_vfolder_hosts"],
    domain_fields["allowed_docker_registries"],
    domain_fields["integration_id"],
)

_default_detail_fields = (
    domain_fields["name"],
    domain_fields["description"],
    domain_fields["is_active"],
    domain_fields["created_at"],
    domain_fields["total_resource_slots"],
    domain_fields["allowed_vfolder_hosts"],
    domain_fields["allowed_docker_registries"],
    domain_fields["integration_id"],
)


class Domain(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that fetches various domain
    information.

    .. note::

      All methods in this function class require your API access key to
      have the *admin* privilege.
    """

    @api_function
    @classmethod
    async def list(
        cls,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        Fetches the list of domains.

        :param fields: Additional per-domain query fields to fetch.
        """
        query = textwrap.dedent(
            """\
            query {
                domains {$fields}
            }
        """
        )
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        data = await api_session.get().Admin._query(query)
        return data["domains"]

    @api_function
    @classmethod
    async def detail(
        cls,
        name: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        """
        Fetch information of a domain with name.

        :param name: Name of the domain to fetch.
        :param fields: Additional per-domain query fields to fetch.
        """
        query = textwrap.dedent(
            """\
            query($name: String) {
                domain(name: $name) {$fields}
            }
        """
        )
        query = query.replace("$fields", " ".join(f.field_ref for f in fields))
        variables = {"name": name}
        data = await api_session.get().Admin._query(query, variables)
        return data["domain"]

    @api_function
    @classmethod
    async def create(
        cls,
        name: str,
        *,
        description: str = "",
        is_active: bool = True,
        total_resource_slots: str | Undefined = undefined,
        vfolder_host_perms: str | Undefined = undefined,  # JSON string
        allowed_docker_registries: Sequence[str] | Undefined = undefined,
        integration_id: str | Undefined = undefined,
        fields: Iterable[FieldSpec | str] | None = None,
    ) -> dict:
        """
        Creates a new domain with the given options.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent(
            """\
            mutation($name: String!, $input: DomainInput!) {
                create_domain(name: $name, props: $input) {
                    ok msg domain {$fields}
                }
            }
        """
        )
        resolved_fields = resolve_fields(fields, domain_fields, (domain_fields["name"],))
        query = query.replace("$fields", " ".join(resolved_fields))
        inputs = {
            "description": description,
            "is_active": is_active,
        }
        set_if_set(inputs, "total_resource_slots", total_resource_slots)
        set_if_set(inputs, "allowed_vfolder_hosts", vfolder_host_perms)
        set_if_set(inputs, "allowed_docker_registries", allowed_docker_registries)
        set_if_set(inputs, "integration_id", integration_id)
        variables = {
            "name": name,
            "input": inputs,
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["create_domain"]

    @api_function
    @classmethod
    async def update(
        cls,
        name: str,
        *,
        new_name: str | Undefined = undefined,
        description: str | Undefined = undefined,
        is_active: bool | Undefined = undefined,
        total_resource_slots: str | Undefined = undefined,
        vfolder_host_perms: str | Undefined = undefined,  # JSON string
        allowed_docker_registries: Sequence[str] | Undefined = undefined,
        integration_id: str | Undefined = undefined,
        fields: Iterable[FieldSpec | str] | None = None,
    ) -> dict:
        """
        Update existing domain.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent(
            """\
            mutation($name: String!, $input: ModifyDomainInput!) {
                modify_domain(name: $name, props: $input) {
                    ok msg
                }
            }
        """
        )
        inputs: dict[str, Any] = {}
        set_if_set(inputs, "name", new_name)
        set_if_set(inputs, "description", description)
        set_if_set(inputs, "is_active", is_active)
        set_if_set(inputs, "total_resource_slots", total_resource_slots)
        set_if_set(inputs, "allowed_vfolder_hosts", vfolder_host_perms)
        set_if_set(inputs, "allowed_docker_registries", allowed_docker_registries)
        set_if_set(inputs, "integration_id", integration_id)
        variables = {
            "name": name,
            "input": inputs,
        }
        data = await api_session.get().Admin._query(query, variables)
        return data["modify_domain"]

    @api_function
    @classmethod
    async def delete(cls, name: str):
        """
        Inactivates an existing domain.
        """
        query = textwrap.dedent(
            """\
            mutation($name: String!) {
                delete_domain(name: $name) {
                    ok msg
                }
            }
        """
        )
        variables = {"name": name}
        data = await api_session.get().Admin._query(query, variables)
        return data["delete_domain"]

    @api_function
    @classmethod
    async def purge(cls, name: str):
        """
        Deletes an existing domain.
        """
        query = textwrap.dedent(
            """\
            mutation($name: String!) {
                purge_domain(name: $name) {
                    ok msg
                }
            }
        """
        )
        variables = {"name": name}
        data = await api_session.get().Admin._query(query, variables)
        return data["purge_domain"]



from typing import Final

KERNEL_CANCELLED: Final[str] = "kernel_cancelled"
KERNEL_CREATING: Final[str] = "kernel_creating"
KERNEL_STARTED: Final[str] = "kernel_started"
KERNEL_TERMINATED: Final[str] = "kernel_terminated"
KERNEL_TERMINATING: Final[str] = "kernel_terminating"
SESSION_FAILURE: Final[str] = "session_failure"
SESSION_STARTED: Final[str] = "session_started"
SESSION_SUCCESS: Final[str] = "session_success"
SESSION_TERMINATED: Final[str] = "session_terminated"
